"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import unittest
import numpy as np
import torchvision
from torch.utils.data.dataloader import DataLoader
from torchvision import transforms

from fuse.data.dataset.dataset_wrapper import FuseDatasetWrapper
from fuse.data.sampler.sampler_balanced_batch import FuseSamplerBalancedBatch
from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict


class FuseSamplerBalancedTestCase(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def setUp(self):
        transform = transforms.Compose([
            transforms.ToTensor(),
            transforms.Normalize((0.1307,), (0.3081,))
        ])
        # Create dataset
        self.torch_dataset = torchvision.datasets.MNIST('/tmp/mnist_test', download=True, train=True, transform=transform)
        pass

    def test_balanced_dataset(self):
        num_classes = 10
        probs = num_classes / 100
        num_samples = 60000

        # wrapping torch dataset
        dataset = FuseDatasetWrapper(name='test', dataset=self.torch_dataset, mapping=('image', 'label'))
        dataset.create()
        print(dataset.summary(statistic_keys=['data.label']))
        sampler = FuseSamplerBalancedBatch(dataset=dataset,
                                           balanced_class_name='data.label',
                                           num_balanced_classes=num_classes,
                                           batch_size=5,
                                           # balanced_class_weights=[1,1,1,1,1,1,1,1,1,1])  # relevant when batch size is % num classes
                                           balanced_class_probs=[probs for i in range(num_classes)])

        labels = np.zeros(num_classes)

        # Create dataloader
        dataloader = DataLoader(dataset=dataset, batch_sampler=sampler, num_workers=0)
        iter1 = iter(dataloader)
        for _ in range(len(dataloader)):
            batch_dict = next(iter1)
            labels_in_batch = FuseUtilsHierarchicalDict.get(batch_dict, 'data.label')
            for label in labels_in_batch:
                labels[label] += 1

        # final balance
        print(labels)
        for idx in range(num_classes):
            sampled = labels[idx] / num_samples
            print(f'Class {idx}: {sampled * 100}% of data')
            self.assertAlmostEqual(sampled, probs, delta=probs * 0.5, msg=f'Unbalanced class {idx}, expected 0.1+-0.05 and got {sampled}')

    def test_unbalanced_dataset(self):
        num_classes = 10
        probs = num_classes / 100

        # wrapping torch dataset
        unbalanced_dataset = FuseDatasetWrapper(name='test', dataset=self.torch_dataset, mapping=('image', 'label'))
        unbalanced_dataset.create()

        samples_to_save = []
        stats = [1000, 200, 200, 200, 300, 500, 700, 800, 900, 1000]
        chosen = np.zeros(10)
        for idx in range(60000):
            label = unbalanced_dataset.get(idx, 'data.label')
            if stats[label] > chosen[label]:
                samples_to_save.append(idx)
                chosen[label] += 1

        unbalanced_dataset.samples_description = samples_to_save

        sampler = FuseSamplerBalancedBatch(dataset=unbalanced_dataset,
                                           balanced_class_name='data.label',
                                           num_balanced_classes=num_classes,
                                           batch_size=5,
                                           # balanced_class_weights=[1,1,1,1,1,1,1,1,1,1])  # relevant when batch size is % num classes
                                           balanced_class_probs=[probs for i in range(num_classes)])

        labels = np.zeros(num_classes)

        # Create dataloader
        dataloader = DataLoader(dataset=unbalanced_dataset, batch_sampler=sampler, num_workers=0)
        iter1 = iter(dataloader)
        num_items = 0
        for _ in range(len(dataloader)):
            batch_dict = next(iter1)
            labels_in_batch = FuseUtilsHierarchicalDict.get(batch_dict, 'data.label')
            for label in labels_in_batch:
                labels[label] += 1
                num_items += 1

        # final balance
        print(labels)
        for idx in range(num_classes):
            sampled = labels[idx] / num_items
            print(f'Class {idx}: {sampled * 100}% of data')
            self.assertAlmostEqual(sampled, probs, delta=probs * 0.5, msg=f'Unbalanced class {idx}, expected 0.1(+-0.05) and got {sampled}')

    def tearDown(self):
        pass


if __name__ == '__main__':
    unittest.main()



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Sequence, Dict, Tuple

import torch

from fuse.models.backbones.backbone_inception_resnet_v2 import FuseBackboneInceptionResnetV2
from fuse.models.heads.head_global_pooling_classifier import FuseHeadGlobalPoolingClassifier
from fuse.utils.utils_hierarchical_dict import FuseUtilsHierarchicalDict


class FuseModelDefault(torch.nn.Module):
    """
    Default Fuse model - convolutional neural network with multiple heads
    """

    def __init__(self,
                 conv_inputs: Tuple[Tuple[str, int], ...] = (('data.input.input_0.tensor', 1),),
                 backbone: torch.nn.Module = FuseBackboneInceptionResnetV2(),
                 heads: Sequence[torch.nn.Module] = (FuseHeadGlobalPoolingClassifier(),)
                 ) -> None:
        """
        Default Fuse model - convolutional neural network with multiple heads
        :param conv_inputs:     batch_dict name for model input and its number of input channels
        :param backbone:        PyTorch backbone module - a convolutional neural network
        :param heads:           Sequence of head modules
        """
        super().__init__()
        self.conv_inputs = conv_inputs
        self.backbone = backbone
        self.add_module('backbone', self.backbone)
        self.heads = torch.nn.ModuleList(heads)
        self.add_module('heads', self.heads)

    def forward(self,
                batch_dict: Dict) -> Dict:
        conv_input = torch.cat([FuseUtilsHierarchicalDict.get(batch_dict, conv_input[0]) for conv_input in self.conv_inputs], 1)
        backbone_features = self.backbone.forward(conv_input)
        FuseUtilsHierarchicalDict.set(batch_dict, 'model.backbone_features', backbone_features)

        for head in self.heads:
            batch_dict = head.forward(batch_dict)

        return batch_dict['model']


if __name__ == '__main__':
    from fuse.models.heads.head_dense_segmentation import FuseHeadDenseSegmentation
    import torch
    import os

    os.environ['CUDA_VISIBLE_DEVICES'] = "0, 1"

    DEVICE = 'cpu'  # 'cuda'
    DATAPARALLEL = False  # True

    model = FuseModelDefault(
        conv_inputs=(('data.input.input_0.tensor', 1),),
        backbone=FuseBackboneInceptionResnetV2(),
        heads=[
            FuseHeadGlobalPoolingClassifier(head_name='head_0',
                                            conv_inputs=[('model.backbone_features', 384)],
                                            post_concat_inputs=None,
                                            num_classes=2),

            FuseHeadDenseSegmentation(head_name='head_1',
                                      conv_inputs=[('model.backbone_features', 384)],
                                      num_classes=2)
        ]
    )

    model = model.to(DEVICE)
    if DATAPARALLEL:
        model = torch.nn.DataParallel(model)

    dummy_data = {'data':
                      {'input':
                           {'input_0': {'tensor': torch.zeros([17, 1, 200, 100]).to(DEVICE), 'metadata': None}},
                       'gt':
                           {'gt_0': torch.zeros([0]).to(DEVICE)}
                       }
                  }

    res = {}
    res['model'] = model.forward(dummy_data)
    print('Forward pass shape - head_0: ', end='')
    print(str(res['model']['logits']['head_0'].shape))

    print('\nForward pass shape - head_1: ', end='')
    print(str(res['model']['logits']['head_1'].shape))

    total_params = sum(p.numel() for p in model.parameters())
    if not DATAPARALLEL:
        backbone_params = sum(p.numel() for p in model._modules['backbone'].parameters())
        print('Backbone params = %d' % backbone_params)
        print('Heads params = %d' % (total_params - backbone_params))

    print('\nTotal params = %d' % total_params)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from abc import ABC, abstractmethod
from typing import Any


class FuseVisualizerBase(ABC):

    @abstractmethod
    def visualize(self, sample: Any, block: bool = True) -> None:
        """
        visualize sample
        :param sample: sample
        :param block: set to False if the process should not be blocked until the plot will be closed
        :return: None
        """
        raise NotImplementedError

    @abstractmethod
    def visualize_aug(self, orig_sample: Any, aug_sample: Any, block: bool = True) -> None:
        """
        Visualise and compare augmented and non-augmented version of the sample
        :param orig_sample: original sample
        :param aug_sample: augmented sample
        :param block: set to False if the process should not be blocked until the plot will be closed
        :return: None
        """
        raise NotImplementedError



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Dict, List


from typing import Dict, Any, List, Iterator, Optional, Union, Sequence, Hashable, Callable
import torch.nn as nn
from torch.optim.optimizer import Optimizer
from torch.utils.data.dataloader import DataLoader

from fuse.losses.loss_base import FuseLossBase
from fuse.metrics.metric_base import FuseMetricBase


class FuseManagerState:
    """
    FuseManagerState contains the current state of the manager.
    """

    def __init__(self) -> None:

        self.output_model_dir: str

        self.net: nn.Module
        self.metrics: Dict[str, FuseMetricBase] = {}
        self.losses: Dict[str, FuseLossBase] = {}
        self.optimizer: Optimizer
        self.lr_scheduler: Any
        self.train_params: Dict = {}
        self.opt_closure: Optional[Callable] = None
        # self.ensemble_nets: Sequence[nn.Module]
        self.num_gpus: int
        self.device: str

        # number of epochs:
        self.num_epochs: int
        self.end_epoch: int
        self.current_epoch: int

        # number of batches in a virtual batch - user defined
        self.virtual_batch_size: int

        # updated throughout run
        self.learning_rate: float

        # best epoch
        # user definitions
        self.best_epoch_source: List[Dict[str, str]]
        self.best_epoch_function: List[str] = []
        self.optimization_function: List[str] = []
        self.on_equal_values: List[str] = []
        self.num_models_to_save: int
        # updated throughout run
        self.best_epoch: List[int]
        self.best_epoch_values: List[Dict] = []

        # checkpoint saving user definitions
        self.gap_between_saving_epochs: int
        self.start_saving_epochs: int


"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import os, glob
import numpy as np
import SimpleITK as sitk
import pydicom
from scipy.ndimage.morphology import binary_dilation


# ========================================================================
# sequences to be read, and the sequence name
SEQ_DICT_PROSTATEX = \
    {
        't2_tse_tra': 'T2',
        't2_tse_tra_Grappa3': 'T2',
        't2_tse_tra_320_p2': 'T2',

        'ep2d-advdiff-3Scan-high bvalue 100': 'b',
        'ep2d-advdiff-3Scan-high bvalue 500': 'b',
        'ep2d-advdiff-3Scan-high bvalue 1400': 'b',
        'ep2d_diff_tra2x2_Noise0_FS_DYNDISTCALC_BVAL': 'b',

        'ep2d_diff_tra_DYNDIST': 'b_mix',
        'ep2d_diff_tra_DYNDIST_MIX': 'b_mix',
        'diffusie-3Scan-4bval_fs': 'b_mix',
        'ep2d_DIFF_tra_b50_500_800_1400_alle_spoelen': 'b_mix',
        'diff tra b 50 500 800 WIP511b alle spoelen': 'b_mix',

        'ep2d_diff_tra_DYNDIST_MIX_ADC': 'ADC',
        'diffusie-3Scan-4bval_fs_ADC': 'ADC',
        'ep2d-advdiff-MDDW-12dir_spair_511b_ADC': 'ADC',
        'ep2d-advdiff-3Scan-4bval_spair_511b_ADC': 'ADC',
        'ep2d_DIFF_tra_b50_500_800_1400_alle_spoelen_ADC': 'ADC',
        'diff tra b 50 500 800 WIP511b alle spoelen_ADC': 'ADC',
        'ADC_S3_1': 'ADC',
        'ep2d_diff_tra_DYNDIST_ADC': 'ADC',

    }

SEQ_TO_USE = ['T2', 'b', 'b_mix', 'ADC', 'ktrans']
SUB_SEQ_TO_USE = ['T2', 'b400', 'b800', 'ADC', 'ktrans']

# sequences with special fix
B_SER_FIX = ['diffusie-3Scan-4bval_fs',
             'ep2d_DIFF_tra_b50_500_800_1400_alle_spoelen',
             'diff tra b 50 500 800 WIP511b alle spoelen']

# patients with special fix
EXP_PATIENTS = ['ProstateX-0191', 'ProstateX-0148', 'ProstateX-0180']

# use T2[-1],b400 [1],b800 [2]
SER_INX_TO_USE = {}
SER_INX_TO_USE['all'] = {'T2': -1, 'b': [1, 2], 'ADC': 0, 'ktrans': 0}
SER_INX_TO_USE['ProstateX-0148'] = {'T2': 1, 'b': [1, 2], 'ADC': 0, 'ktrans': 0}
SER_INX_TO_USE['ProstateX-0191'] = {'T2': -1, 'b': [0, 0], 'ADC': 0, 'ktrans': 0}
SER_INX_TO_USE['ProstateX-0180'] = {'T2': -1, 'b': [1, 2], 'ADC': 0, 'ktrans': 0}

# ==========================================================================
def create_resample(vol_ref, interpolation,size,spacing):
    if interpolation == 'linear':
        interpolator = sitk.sitkLinear
    elif interpolation == 'nn':
        interpolator = sitk.sitkNearestNeighbor
    elif interpolation == 'bspline':
        interpolator = sitk.sitkBSpline

    resample = sitk.ResampleImageFilter()
    resample.SetReferenceImage(vol_ref)
    resample.SetOutputSpacing(spacing)
    resample.SetInterpolator(interpolator)
    resample.SetSize(size)
    return resample

class PROSTATEX3DVolume:
    def __init__(self,
                 patient_id: str=None,
                 imgs_path: str=None,
                 ktrans_data_path: str=None,
                 array: np.ndarray = None,
                 hdf5_filename: str = None,
                 verbose: bool = True,
                 ):

        self._patient_id = patient_id
        self._imgs_path = imgs_path
        self._ktrans_data_path = ktrans_data_path
        self._array = array
        self._hdf5_filename = hdf5_filename
        self._verbose = verbose
        if array is None and imgs_path is None and hdf5_filename is None:
            print('ERROR link to volume OR path to images must be provided')
            assert False, 'ERROR while creating CAP3DVolume: no input was provided'

    # ========================================================================

    def extract_stk_vol(self,img_path, img_list=[], reverse_order=False, is_path=True):

            stk_vols = []
            if is_path:
                vol = sitk.ReadImage(img_path)
                stk_vols.append(vol)
                return stk_vols

            series_reader = sitk.ImageSeriesReader()

            if img_list == []:
                img_list = [series_reader.GetGDCMSeriesFileNames(img_path)]

            for n, imgs_names in enumerate(img_list):
                if img_path not in img_list[0][0]:
                    imgs_names = [os.path.join(img_path, n) for n in imgs_names]
                dicom_names = imgs_names[::-1] if reverse_order else imgs_names
                series_reader.SetFileNames(dicom_names)
                imgs = series_reader.Execute()
                stk_vols.append(imgs)

            return stk_vols

    # ========================================================================

    def sort_dicom_by_dicom_field(self,dcm_files, dicom_field=(0x19, 0x100c)):
        dcm_values = {}
        for dcm in dcm_files:
            dcm_ds = pydicom.dcmread(dcm)

            val = int(dcm_ds[dicom_field].value)
            if val not in dcm_values:
                dcm_values[val] = []
            dcm_values[val].append(os.path.split(dcm)[-1])
        sorted_names_list = [v for k, v in dcm_values.items()]
        return sorted_names_list

    # ========================================================================
    def extract_prostatex_stk_vol(self):


        ktrans_path = os.path.join(self._ktrans_data_path, self._patient_id)

        if self._verbose: print('Patient ID: %s' % (self._patient_id))

        # ------------------------
        # images dict and sequences description dict

        vols_dict = {k: [] for k in SEQ_TO_USE}
        sequences_dict = {k: [] for k in SEQ_TO_USE}


        for img_path in os.listdir(self._imgs_path):
            try:
                full_path = os.path.join(self._imgs_path, img_path)
                dcm_files = glob.glob(os.path.join(full_path, '*.dcm'))
                series_desc = pydicom.dcmread(dcm_files[0]).SeriesDescription

                #------------------------
                # print series description
                series_desc_general = SEQ_DICT_PROSTATEX[series_desc] \
                    if series_desc in SEQ_DICT_PROSTATEX else 'UNKNOWN'
                if self._verbose: print('\t- Series description:',' %s (%s)' % (series_desc, series_desc_general))

                #------------------------
                # ignore UNKNOWN series
                if series_desc not in SEQ_DICT_PROSTATEX or \
                        SEQ_DICT_PROSTATEX[series_desc] not in SEQ_TO_USE:
                    continue

                #------------------------
                # b-series - sorting images by b-value

                if SEQ_DICT_PROSTATEX[series_desc] == 'b_mix':

                    sorted_dicom_names = self.sort_dicom_by_dicom_field(dcm_files, dicom_field=(0x19, 0x100c))
                    stk_vols = self.extract_stk_vol(full_path, img_list=sorted_dicom_names, reverse_order=True, is_path=False)
                #------------------------
                # general case
                else:
                    stk_vols = self.extract_stk_vol(full_path, img_list=[], reverse_order=False, is_path=False)

                #------------------------
                # volume dictionary

                if SEQ_DICT_PROSTATEX[series_desc] == 'b_mix':
                    vols_dict['b'] += stk_vols
                    sequences_dict['b'] += [series_desc]
                else:
                    vols_dict[SEQ_DICT_PROSTATEX[series_desc]] += stk_vols
                    sequences_dict[SEQ_DICT_PROSTATEX[series_desc]] += [series_desc]

            except Exception as e:
                print(e)

        # Read ktrans image
        try:

            mhd_path = glob.glob(os.path.join(ktrans_path, '*.mhd'))[0]
            print('\t- Reading: %s (%s) (%s)' % (os.path.split(mhd_path)[-1], 'Ktrans', 'ktrans'))
            stk_vols = self.extract_stk_vol(mhd_path, img_list=[], reverse_order=False, is_path=True)
            vols_dict['ktrans'] = stk_vols
            sequences_dict['ktrans'] = [ktrans_path]


        except Exception as e:
            print(e)

        if 'b_mix' in vols_dict.keys():
            vols_dict.pop('b_mix')
            sequences_dict.pop('b_mix')
        return vols_dict, sequences_dict

    # ========================================================================
    def read_prostatex_sequences_patient(self):

        def get_zeros_vol(vol):

            if vol.GetNumberOfComponentsPerPixel() > 1:
                ref_zeros_vol = sitk.VectorIndexSelectionCast(vol, 0)
            else:
                ref_zeros_vol = vol
            zeros_vol = np.zeros_like(sitk.GetArrayFromImage(ref_zeros_vol))
            zeros_vol = sitk.GetImageFromArray(zeros_vol)
            zeros_vol.CopyInformation(ref_zeros_vol)
            return zeros_vol

        def stack_rel_vol_in_list(vols,series_inx_to_use,seq):
            vols_list = []
            for s, v0 in vols.items():
                vol_inx_to_use = series_inx_to_use['all'][s]

                if self._patient_id in EXP_PATIENTS:
                    vol_inx_to_use = series_inx_to_use[self._patient_id][s]

                if isinstance(vol_inx_to_use,list):
                    for inx in vol_inx_to_use:
                        v = v0[inx]
                        if len(v):
                            vols_list.append(v)
                        else:
                            vols_list.append(get_zeros_vol(vols_list[0]))
                            if self._verbose: print('\n - problem with reading %s volume!' % s)

                else:
                    v = v0[vol_inx_to_use]
                    if len(v):
                        vols_list.append(v)
                    else:
                        vols_list.append(get_zeros_vol(vols_list[0]))
                        if self._verbose: print('\n - problem with reading %s volume!' % s)

            if ('b' in seq_info.keys()):
                if (seq_info['b'][0] in B_SER_FIX):
                    vols_list[seq.index('b800')].CopyInformation(vols_list[seq.index('ADC')])
                    vols_list[seq.index('b400')].CopyInformation(vols_list[seq.index('ADC')])

            if len(vols_list) != len(seq):
                raise ValueError('Expected %d image modalities, found %d' % (len(seq), len(vols_list)))

            return vols_list


        # ------------------------
        # extract stk vol list per sequence
        vols_dict, seq_info = self.extract_prostatex_stk_vol()

        # ------------------------
        # stack volumes by  seq order,
        # keep only vol as defined in series_inx_to_use
        vols_list = stack_rel_vol_in_list(vols_dict, SER_INX_TO_USE,SUB_SEQ_TO_USE)

        return vols_list


    # ========================================================================
    def apply_rescaling(self,img, thres=(1.0, 99.0), method='noclip'):

        eps = 0.000001

        def rescale_single_channel_image(img):
            # Deal with negative values first
            min_value = np.min(img)
            if min_value < 0:
                img -= min_value
            if method == 'clip':
                val_l, val_h = np.percentile(img, thres)
                img2 = img
                img2[img < val_l] = val_l
                img2[img > val_h] = val_h
                img2 = (img2.astype(np.float32) - val_l) / (val_h - val_l + eps)
            elif method == 'mean':
                img2 = img / max(np.mean(img), 1)
            elif method == 'median':
                img2 = img / max(np.median(img), 1)
            elif method == 'noclip':
                val_l, val_h = np.percentile(img, thres)
                img2 = img
                img2 = (img2.astype(np.float32) - val_l) / (val_h - val_l + eps)
            else:
                img2 = img
            return img2

        # fix outlier image values
        img[np.isnan(img)] = 0
        # img[img < 0] = 0
        # Process each channel independently
        if len(img.shape) == 4:
            for i in range(img.shape[-1]):
                img[..., i] = rescale_single_channel_image(img[..., i])
        else:
            img = rescale_single_channel_image(img)

        return img


    # ========================================================================
    def apply_resampling(self,img, mask, spacing=(0.5, 0.5, 3), size=(160, 160, 32),
                             transform=None, interpolation='bspline',
                             label_interpolator=sitk.sitkLabelGaussian,
                             ):

        ref = img if img != [] else mask
        size = [int(s) for s in size]
        resample = create_resample(ref, interpolation, size=size, spacing=spacing)

        if ~(transform is None):
            resample.SetTransform(transform)
        img_r = resample.Execute(img)

        resample.SetInterpolator(label_interpolator)
        mask_r = resample.Execute(mask)


        return img_r, mask_r


    # ========================================================================
    def preprocess_sequences(self,vols,interpolation='bspline', reference_inx=0):

        # ------------------------
        # casting to float32
        vols = [sitk.Cast(im, sitk.sitkFloat32) for im in vols]

        # ------------------------
        # create resampling oparator based on ref vol

        # define the reference volume
        vol_ref = vols[reference_inx]
        other_inx = list(set(range(0, len(vols))) - set([reference_inx]))

        resample = create_resample(vol_ref, interpolation,size=vol_ref.GetSize() ,spacing=vol_ref.GetSpacing())

        vols_res = []
        for i,vol in enumerate(vols):
            if i in other_inx:
                vol_res = resample.Execute(vol)
            else:
                vol_res = vol_ref
            vols_res.append(vol_res)

        # ------------------------
        # stack sequences in 4D sitk

        vol_arr = [sitk.GetArrayFromImage(vol) for vol in (vols_res)]
        # print('1'+str(len(vol_arr)))
        vol_final = np.stack(vol_arr, axis=-1)
        # print('2'+str(vol_final.shape))
        vol_final_sitk = sitk.GetImageFromArray(vol_final, isVector=True)
        # print('3'+str(vol_final_sitk.GetSize()))
        vol_final_sitk.CopyInformation(vol_ref)

        # ------------------------
        # rescale intensity

        vol_backup = sitk.Image(vol_final_sitk)
        vol_array = sitk.GetArrayFromImage(vol_final_sitk)
        if len(vol_array.shape)<4:
            vol_array= vol_array[:,:,:,np.newaxis]
        # print('4' + str(vol_array.shape))
        vol_array = self.apply_rescaling(vol_array)
        vol_final = sitk.GetImageFromArray(vol_array, isVector=True)
        # print('5' + str(vol_final.GetSize()))
        vol_final.CopyInformation(vol_backup)
        vol_final = sitk.Image(vol_final)

        return vol_final


    # ========================================================================
    def crop_lesion_vol(self,vol, position, ref, size=(160, 160, 32), spacing=(1, 1, 3), center_slice=None):

        def get_lesion_mask(position, ref):
            mask = np.zeros_like(sitk.GetArrayViewFromImage(ref), dtype=np.uint8)

            coords = np.round(position[::-1]).astype(np.int)
            mask[coords[0], coords[1], coords[2]] = 1
            mask = binary_dilation(mask, np.ones((3, 5, 5))) + 0
            mask_sitk = sitk.GetImageFromArray(mask)
            mask_sitk.CopyInformation(ref)

            return mask_sitk

        mask = get_lesion_mask(position, ref)

        vol.SetOrigin((0,) * 3)
        mask.SetOrigin((0,) * 3)
        vol.SetDirection(np.eye(3).flatten())
        mask.SetDirection(np.eye(3).flatten())

        ma_centroid = mask > 0.5
        label_analysis_filer = sitk.LabelShapeStatisticsImageFilter()
        label_analysis_filer.Execute(ma_centroid)
        centroid = label_analysis_filer.GetCentroid(1)
        offset_correction = np.array(size) * np.array(spacing)/2
        corrected_centroid = np.array(centroid)
        corrected_centroid[2] = center_slice * np.array(spacing[2])
        offset = corrected_centroid - np.array(offset_correction)

        translation = sitk.TranslationTransform(3, offset)
        img, mask = self.apply_resampling(vol, mask, spacing=spacing, size=size, transform=translation)

        return img, mask


if __name__ == "__main__":
    path_to_db = '/gpfs/haifa/projects/m/msieve_dev3/usr/Tal/my_research/virtual_biopsy/prostate/experiments/V1/'
    dataset = 'prostate_x'
    if dataset=='tcia':
        # for TCIA-Prostate-MRI-US-Biopsy
        prostate_data_path = '/projects/msieve/MedicalSieve/PatientData/TCIA-Prostate-MRI-US-Biopsy/manifest-1599764098812/'
        masks_data_path = prostate_data_path + '/STLs/'
        sample = ('18042021', 'train', 'Prostate-MRI-US-Biopsy-0007', 'pred')

        a = PROSTATEX3DVolume(path_to_db,prostate_data_path,masks_data_path,dataset,(16, 120, 120))
        samples = a.__call__(sample)
        a = 1
    elif dataset=='prostate_x':
        EXP_LESION_INX_PATIENTS = ['ProstateX-0005', 'ProstateX-0105', 'ProstateX-0154']
        EXP_LESION_REMOVE_PATIENTS = ['ProstateX-0025']

    # for ProstateX

        path_imgs = '/projects/msieve/MedicalSieve/PatientData/ProstateX/manifest-A3Y4AE4o5818678569166032044//PROSTATEx/ProstateX-0148/03-12-2012-MR prostaat kanker detectie WDSmc MCAPRODETW-33123'
        patient_id = 'ProstateX-0148'
        path_ktrans = '/projects/msieve/MedicalSieve/PatientData/ProstateX/manifest-A3Y4AE4o5818678569166032044//ProstateXKtrains-train-fixed/'

        vol3D = PROSTATEX3DVolume(patient_id=patient_id,imgs_path=path_imgs,ktrans_data_path=path_ktrans)

        reference_inx = 0
        lsn_shape = (12, 32, 32)
        spacing = (0.5, 0.5, 3)

        vol_list = vol3D.read_prostatex_sequences_patient()
        vol = vol3D.preprocess_sequences(vol_list, reference_inx=reference_inx)

        positions_vol = np.array([146.84003826 ,207.70670297,   4.00001143]) #T2 coord
        vol_cropped, mask_cropped = vol3D.crop_lesion_vol(vol, positions_vol, vol_list[reference_inx],
                                                          center_slice=positions_vol[2],
                                                          size=(lsn_shape[2], lsn_shape[1], lsn_shape[0]), spacing=spacing)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import numpy as np
from sklearn import metrics

from fuse.metrics.classification.metric_default import FuseMetricDefault


class FuseMetricAUC(FuseMetricDefault):
    """
    Area under the receiver operating characteristic curve
    Multi class version
    """

    def __init__(self,
                 **kwargs) -> None:
        """
        See FuseMetricDefault
        """
        super().__init__(metrics.roc_auc_score, **kwargs)


class FuseMetricAUCPR(FuseMetricDefault):
    """
    Area under the precision recall curve
    Multi class version
    """

    def __init__(self,
                 **kwargs) -> None:
        """
        See FuseMetricDefault
        """
        super().__init__(aucpr, **kwargs)


def aucpr(y_true:np.ndarray, y_score:np.ndarray, sample_weight:np.ndarray):
    precision, recall, _ = metrics.precision_recall_curve(y_true, y_score,
                                                          sample_weight=sample_weight)
    return metrics.auc(recall, precision)


if __name__ == '__main__':
    data = {'preds': np.array([[0.8, 0.1, 0.1],
                               [0.5, 0.3, 0.2],
                               [0.6, 0.3, 0.1],
                               [0.6, 0.1, 0.3],
                               [0.7, 0.2, 0.1],
                               [0.3, 0.2, 0.5],
                               [0.1, 0.2, 0.7],
                               [0.2, 0.6, 0.2],
                               [0.3, 0.3, 0.4],
                               [0.7, 0.2, 0.1]]),
            'targets': np.array([0, 0, 0, 0, 0, 1, 1, 1, 2, 2]),
            'weights': np.array([0.03, 0.12, 0.05, 0.52, 0.23, 0.72, 0.13, 0.113, 0.84, 0.09])}

    metric_dict = {"auc": FuseMetricAUC(pred_name='preds', target_name='targets', use_sample_weights=True,
                                        sample_weight_name='weights'),
                   "aucpr": FuseMetricAUCPR(pred_name='preds', target_name='targets', use_sample_weights=True,
                                            sample_weight_name='weights')}
    for metric_name, metric in metric_dict.items():
        metric.collect(data)
        res = metric.process()
        print("use_sample_weights = True")
        for k, v in res.items():
            print(k, '\t', v)

        print("use_sample_weights = False")
        metric.use_sample_weights = False

        res = metric.process()

        for k, v in res.items():
            print(k, '\t', v)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import logging
from typing import List


class FuseProstateXTask():
    tasks = {}
    def __init__(self, task_name: str, version: int):
        self._task_name, self._task_version, self._task_mapping, self._task_class_names = \
            self.get_task(task_name, version)

    def name(self):
        return self._task_name + "_"  + str(self._task_version)

    def class_names(self):
        return self._task_class_names

    def num_classes(self):
        return len(self._task_class_names)

    def mapping(self):
        return self._task_mapping

    @classmethod
    def register(cls, name: str, version: int, mapping: List, class_names: List[str]):
        key = (name, version)
        assert key not in cls.tasks
        cls.tasks[key] = (name, version, mapping, class_names)

    @classmethod
    def get_task(cls, task_name: str, version: int):
        key = (task_name, version)
        if key not in cls.tasks:
            msg = f'Task not found - list of tasks: {list(cls.tasks.keys())}'
            logging.getLogger('Fuse').error(msg)
            raise Exception(msg)

        return cls.tasks[key]



#DO NOT CHANGE TASKS!!!!
GLEASON_SCORE = ['HIGH','LOW','BENIGN']
GLEASON_SCORE_VER_0 = [['HIGH'], ['LOW'],['BENIGN']],
CLINSIG_VER_0 = [['HIGH'], ['LOW']],


FuseProstateXTask.register('gleason_score', 0, GLEASON_SCORE_VER_0, ['HIGH','LOW','BENIGN'])
FuseProstateXTask.register('ClinSig', 0, CLINSIG_VER_0, ['HIGH','LOW'])
if __name__ == '__main__':
    mp_task = FuseProstateXTask('gleason_score', 0)
    print(mp_task.name())
    print(mp_task.class_names())
    print(len(mp_task.class_names()))
    print(mp_task.mapping())


"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import os

from fuse.utils.utils_debug import FuseUtilsDebug
from fuse.utils.utils_gpu import FuseUtilsGPU

os.environ['skip_broker'] = '1'

import logging

import torch.optim as optim
from torch.utils.data.dataloader import DataLoader

from fuse.utils.utils_logger import fuse_logger_start

from fuse.data.sampler.sampler_balanced_batch import FuseSamplerBalancedBatch
from fuse.data.visualizer.visualizer_default import FuseVisualizerDefault
from fuse.data.augmentor.augmentor_default import FuseAugmentorDefault
from fuse.data.dataset.dataset_default import FuseDatasetDefault

from fuse.models.model_default import FuseModelDefault

from fuse.managers.callbacks.callback_tensorboard import FuseTensorboardCallback
from fuse.managers.callbacks.callback_metric_statistics import FuseMetricStatisticsCallback
from fuse.managers.callbacks.callback_time_statistics import FuseTimeStatisticsCallback
from fuse.managers.manager_default import FuseManagerDefault

from fuse.analyzer.analyzer_default import FuseAnalyzerDefault

###########################################################################################################
# Fuse
# Example of a training template:
#
# Training template contains 6 fundamental building blocks:
# (1) Data
# (2) Model
# (3) Losses
# (4) Metrics
# (5) Callbacks
# (6) Manager
#
# The template will create each of those building and will eventually run Manager.train()
#
# Terminology:
# 1.batch_dict -
#   Dictionary that aggregates data and results during the batch cycle.
#   The resulted batch is hierarchical dictionary including 3 branches: 'data', 'model', 'losses'
#   batch_dict allows us to decouple the training components.
#   For example 'classification_loss' code don't directly interact with the data or model,
#   Instead a string name, which is a key in batch dict, will be specified for both the:
#   (1) classification label: 'data.gt.global.label' (2) model logits predictions: 'model.head_0.logits'
# 2.epoch_result -
#   A dictionary created by the manager and includes the losses and metrics value
#   as defined within the template and calculated for the specific epoch
# 3.Fuse base classes - Fuse*Base -
#   Abstract classes of the object forming together Fuse framework .
# 4.Fuse default classes - Fuse*Default -
#   A default generic implementation of the equivalent base class.
#   Those generic implementation will be useful for most common use cases.
#   Alternative implementations could be implemented for the special cases.
# 5.sample_descriptor -
#   Unique ID representing sample
###########################################################################################################

##########################################
# Debug modes
##########################################
mode = 'default'  # Options: 'default', 'fast', 'debug', 'verbose', 'user'. See details in FuseUtilsDebug
debug = FuseUtilsDebug(mode)

##########################################
# Output Paths
##########################################
PATHS = {'model_dir': 'TODO',
         'force_reset_model_dir': True,  # If True will reset model dir automatically - otherwise will prompt 'are you sure' message.
         'cache_dir': 'TODO',
         'inference_dir': 'TODO'}

##########################################
# Train Common Params
##########################################
# ============
# Data
# ============
TRAIN_COMMON_PARAMS = {}
TRAIN_COMMON_PARAMS['data.batch_size'] = 2
TRAIN_COMMON_PARAMS['data.train_num_workers'] = 8
TRAIN_COMMON_PARAMS['data.validation_num_workers'] = 8
TRAIN_COMMON_PARAMS['data.augmentation_pipeline'] = [
    # TODO: define the augmentation pipeline here
    # Fuse TIP: Use as a reference the simple augmentation pipeline written in Fuse.data.augmentor.augmentor_toolbox.aug_image_default_pipeline
]
# ===============
# Manager - Train
# ===============
TRAIN_COMMON_PARAMS['manager.train_params'] = {
    'num_epochs': 100,
    'virtual_batch_size': 1,  # number of batches in one virtual batch
    'start_saving_epochs': 10,  # first epoch to start saving checkpoints from
    'gap_between_saving_epochs': 5,  # number of epochs between saved checkpoint
}
TRAIN_COMMON_PARAMS['manager.best_epoch_source'] = {
    'source': 'TODO',  # can be any key from 'epoch_results' (either metrics or losses result)
    'optimization': 'max',  # can be either min/max
    'on_equal_values': 'better',
    # can be either better/worse - whether to consider best epoch when values are equal
}
TRAIN_COMMON_PARAMS['manager.learning_rate'] = 1e-4
TRAIN_COMMON_PARAMS['manager.weight_decay'] = 0.001
TRAIN_COMMON_PARAMS['manager.resume_checkpoint_filename'] = None  # if not None, will try to load the checkpoint


#################################
# Train Template
#################################
def train_template(paths: dict, train_common_params: dict):
    # ==============================================================================
    # Logger
    #   - output log automatically to three destinations:
    #     (1) console (2) file - copy of the console (3) verbose file - used for debug
    #   - save a copy of the template file
    # ==============================================================================
    fuse_logger_start(output_path=paths['model_dir'], console_verbose_level=logging.INFO)
    lgr = logging.getLogger('Fuse')
    lgr.info('Fuse Train', {'attrs': ['bold', 'underline']})

    lgr.info(f'model_dir={paths["model_dir"]}', {'color': 'magenta'})
    lgr.info(f'cache_dir={paths["cache_dir"]}', {'color': 'magenta'})

    # ==============================================================================
    # Data
    #   Build dataloaders (torch.utils.data.DataLoader) for both train and validation.
    #   Using Fuse generic components:
    #
    #   (1) Data Source - FuseDataSourceBase -
    #       Class providing list of sample descriptors
    #
    #   (2) Processor - FuseProcessorBase -
    #       Group of classes extracting sample data given the sample descriptor.
    #       We divide the processor to two groups (1) 'input' (2) 'gt'.
    #       Both groups will be used for training, but only the 'input' group will be used for 'inference'
    #       The input processors data will be aggregated in batch_dict['data.input.<processor name>.*']
    #       The gt processors data will be aggregated in batch_dict['data.gt.<processor name>.*']
    #       Available Processor classes are:
    #       FuseProcessorCSV -  Reads CSV file, call() returns a sample as dict
    #       FuseProcessorDataFrame -  Reads either data frame or pickled data frame, call() returns a sample as dict
    #       FuseProcessorDataFrameWithGT - Reads either data frame or pickled data frame, call() returns a gt tensor value as dict
    #       FuseProcessorRand - Processor generating random ground truth - useful for testing and sanity check
    #
    #   (3) Dataset - FuseDatasetBase -
    #       Extended pytorch Dataset class - providing additional functionality such as caching, filtering, visualizing.
    #       Available Dataset classes are:
    #       FuseDatasetDefault - generic default implementation
    #       FuseDatasetGenerator - to be used when generating simple samples at once (e.g., patches of a single image)
    #       FuseDatasetWrapper - wraps Pytorch's dataset, converts each sample into a dict.
    #
    #   (4) Augmentor - FuseAugmentorBase -
    #       Optional class applying the augmentation
    #       See FuseAugmentorDefault for default generic implementation of augmentor. It is aimed to be used by most experiments.
    #       See fuse.data.augmentor.augmentor_toolbox.py for implemented augmentation functions to be used in the pipeline.
    #
    #   (5) Visualizer - FuseVisualizerBase -
    #       Optional class visualizing the data before and after augmentations
    #       Available visualizers:
    #       FuseVisualizerDefault - Visualizer for data including single 2D image with optional mask
    #       Fuse3DVisualizerDefault - Visualizer for data including 3D volume with optional mask
    #       FuseVisualizerImageAnalysis - Visualizer for producing analysis of an image
    #
    #   (6) Sampler - implementing 'torch.utils.data.sampler' -
    #       Class retrieving list of samples to use for each batch
    #       Available Sampler;
    #       FuseSamplerBalancedBatch - balances data per batch. Supports balancing of classes by weights/probabilities.
    # ==============================================================================
    #### Train Data

    lgr.info(f'Train Data:', {'attrs': 'bold'})

    ## Create data source:
    # TODO: Create instance of FuseDataSourceBase
    # Reference: FuseMGDataSource
    train_data_source = None

    ## Create data processors:
    # TODO - Create instances of FuseProcessorBase and add to the dictionaries below
    # Reference: FuseProcessorDataFrame, FuseProcessorCSV, DatasetProcessor (for pytorch data)
    input_processors = {}
    gt_processors = {}

    ## Create data augmentation (optional)
    augmentor = FuseAugmentorDefault(augmentation_pipeline=train_common_params['data.augmentation_pipeline'])

    # Create visualizer (optional)
    # TODO - Either use the default visualizer or an alternative one
    visualiser = FuseVisualizerDefault(image_name='TODO', label_name='TODO')

    # Create dataset
    # Fuse TIP: If it's more convenient to generate few samples at once, use FuseDatasetGenerator
    train_dataset = FuseDatasetDefault(cache_dest=paths['cache_dir'],
                                       data_source=train_data_source,
                                       input_processors=input_processors,
                                       gt_processors=gt_processors,
                                       augmentor=augmentor,
                                       visualizer=visualiser)

    lgr.info(f'- Load and cache data:')
    train_dataset.create()
    lgr.info(f'- Load and cache data: Done')

    ## Fuse TIPs:
    # 1. Get to know the resulted data structure train_dataset[0]
    # 2. Visualize here the input to the network using 'train_dataset.visualize(sample_index)
    # 3. Compare augmented case to non-augmented case using 'train_dataset.visualize_augmentation(sample_index)'
    # 4. Review the summary generated by train_dataset.summary() including basic statistics

    ## Create sampler
    # Fuse TIPs:
    # 1. You don't have to balance according the classification labels, any categorical value will do.
    #    Use balanced_class_name to select the categorical value
    # 2. You don't have to equally balance between the classes.
    #    Use balanced_class_weights to specify the number of required samples in a batch per each class
    lgr.info(f'- Create sampler:')
    sampler = FuseSamplerBalancedBatch(dataset=train_dataset,
                                       balanced_class_name='TODO',
                                       num_balanced_classes='TODO',
                                       batch_size=train_common_params['data.batch_size'],
                                       balanced_class_weights=None)

    lgr.info(f'- Create sampler: Done')

    ## Create dataloader
    train_dataloader = DataLoader(dataset=train_dataset,
                                  shuffle=False, drop_last=False,
                                  batch_sampler=sampler,
                                  collate_fn=train_dataset.collate_fn,
                                  num_workers=train_common_params['data.train_num_workers'])
    lgr.info(f'Train Data: Done', {'attrs': 'bold'})

    #### Validation data
    lgr.info(f'Validation Data:', {'attrs': 'bold'})

    ## Create data source
    # TODO: Create instance of FuseDataSourceBase
    # Reference: FuseDataSourceDefault
    validation_data_source = 'TODO'

    ## Create dataset
    # Fuse TIP: If it's more convenient to generate few samples at once, use FuseDatasetGenerator
    validation_dataset = FuseDatasetDefault(cache_dest=paths['cache_dir'],
                                            data_source=validation_data_source,
                                            input_processors=input_processors,
                                            gt_processors=gt_processors,
                                            augmentor=None,
                                            visualizer=visualiser)

    lgr.info(f'- Load and cache data:')
    validation_dataset.create()
    lgr.info(f'- Load and cache data: Done')

    ## Create dataloader
    validation_dataloader = DataLoader(dataset=validation_dataset,
                                       shuffle=False,
                                       drop_last=False,
                                       batch_sampler=None,
                                       batch_size=train_common_params['data.batch_size'],
                                       num_workers=train_common_params['data.validation_num_workers'],
                                       collate_fn=validation_dataset.collate_fn)
    lgr.info(f'Validation Data: Done', {'attrs': 'bold'})

    # ===================================================================================================================
    # Model
    #   Build a model (torch.nn.Module) using generic Fuse componnets:
    #   1. FuseModelDefault - generic component supporting single backbone with multiple heads
    #   2. FuseBackbone - simple backbone model
    #   3. FuseHead* - generic head implementations
    #   The model outputs will be aggregated in batch_dict['model.*']
    #   Each head output will be aggregated in batch_dict['model.<head name>.*']
    #
    #   Additional implemented models:
    #   * FuseModelEnsemble - runs several sub-modules sequentially
    #   * FuseModelMultistream - convolutional neural network with multiple processing streams and multiple heads
    #   *
    # ===================================================================================================================
    lgr.info('Model:', {'attrs': 'bold'})
    # TODO - define / create a model
    model = FuseModelDefault(
        conv_inputs=(('data.input.input_0.tensor', 1),),
        backbone='TODO',  # Reference: FuseBackboneInceptionResnetV2
        heads=['TODO']  # References: FuseHeadGlobalPoolingClassifier, FuseHeadDenseSegmentation
    )
    lgr.info('Model: Done', {'attrs': 'bold'})

    # ==========================================================================================================================================
    #   Loss
    #   Dictionary of loss elements each element is a sub-class of FuseLossBase
    #   The total loss will be the weighted sum of all the elements.
    #   Each element output loss will be aggregated in batch_dict['losses.<loss name>']
    #   The average batch loss per epoch will be included in epoch_result['losses.<loss name>'],
    #   and the total loss in epoch_result['losses.total_loss']
    #   The 'best_epoch_source', used to save the best model could be based on one of this losses.
    #   Available Losses:
    #   FuseLossDefault - wraps a PyTorch loss function with a Fuse api.
    #   FuseLossSegmentationCrossEntropy - calculates cross entropy loss per location ("dense") of a class activation map ("segmentation")
    #
    # ==========================================================================================================================================
    losses = {
        # TODO add losses here (instances of FuseLossBase)
    }

    # =========================================================================================================
    # Metrics
    # Dictionary of metric elements. Each element is a sub-class of FuseMetricBase
    # The metrics will be calculated per epoch for both the validation and train.
    # The results will be included  in epoch_result['metrics.<metric name>']
    # The 'best_epoch_source', used to save the best model could be based on one of this metrics.
    # Available Metrics:
    # FuseMetricAccuracy - basic accuracy
    # FuseMetricAUC - Area under receiver operating characteristic curve
    # FuseMetricBoundingBoxes - Bounding boxes metric
    # FuseMetricConfidenceInterval - Wrapper Metric to compute the confidence interval of another metric
    # FuseMetricConfusionMatrix - Confusion matrix
    # FuseMetricPartialAUC - Partial area under receiver operating characteristic curve
    # FuseMetricScoreMap - Segmentation metric
    #
    # =========================================================================================================
    metrics = {
        # TODO add metrics here (instances of FuseMetricBase)
    }

    # ==========================================================================================================
    #  Callbacks
    #  Callbacks are sub-classes of FuseCallbackBase.
    #  A callback is an object that can perform actions at various stages of training,
    #  In each stage it allows to manipulate either the data, batch_dict or epoch_results.
    # ==========================================================================================================
    callbacks = [
        # Fuse TIPs: add additional callbacks here
        # default callbacks
        FuseTensorboardCallback(model_dir=paths['model_dir']),  # save statstics for tensorboard
        FuseMetricStatisticsCallback(output_path=paths['model_dir'] + "/metrics.csv"),  # save statisticsin a csv file
        FuseTimeStatisticsCallback(num_epochs=train_common_params['manager.train_params']['num_epochs'], load_expected_part=0.1)  # time profiler
    ]

    # =====================================================================================
    #  Manager - Train
    #  Create a manager, training objects and run a training process.
    # =====================================================================================
    lgr.info('Train:', {'attrs': 'bold'})

    # create optimizer
    optimizer = optim.Adam(model.parameters(), lr=train_common_params['manager.learning_rate'],
                           weight_decay=train_common_params['manager.weight_decay'])

    # create scheduler
    scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer)

    # train from scratch
    manager = FuseManagerDefault(output_model_dir=paths['model_dir'], force_reset=paths['force_reset_model_dir'])
    # Providing the objects required for the training process.
    manager.set_objects(net=model,
                        optimizer=optimizer,
                        losses=losses,
                        metrics=metrics,
                        best_epoch_source=train_common_params['manager.best_epoch_source'],
                        lr_scheduler=scheduler,
                        callbacks=callbacks,
                        train_params=train_common_params['manager.train_params'])

    ## Continue training
    if train_common_params['manager.resume_checkpoint_filename'] is not None:
        # Loading the checkpoint including model weights, learning rate, and epoch_index.
        manager.load_checkpoint(checkpoint=train_common_params['manager.resume_checkpoint_filename'], mode='train')

    # Start training
    manager.train(train_dataloader=train_dataloader,
                  validation_dataloader=validation_dataloader)

    lgr.info('Train: Done', {'attrs': 'bold'})


######################################
# Inference Common Params
######################################
INFER_COMMON_PARAMS = {}
INFER_COMMON_PARAMS['infer_filename'] = os.path.join(PATHS['inference_dir'], 'validation_set_infer.pickle')
INFER_COMMON_PARAMS['checkpoint'] = 'best'  # Fuse TIP: possible values are 'best', 'last' or epoch_index.


######################################
# Inference Template
######################################
def infer_template(paths: dict, infer_common_params: dict):
    #### Logger
    fuse_logger_start(output_path=paths['inference_dir'], console_verbose_level=logging.INFO)
    lgr = logging.getLogger('Fuse')
    lgr.info('Fuse Inference', {'attrs': ['bold', 'underline']})
    lgr.info(f'infer_filename={infer_common_params["infer_filename"]}', {'color': 'magenta'})

    #### create infer datasource
    # TODO: Create instance of FuseDataSourceBase
    # Reference: FuseDataSourceDefault
    infer_data_source = 'TODO'
    lgr.info(f'experiment={infer_common_params["experiment_filename"]}', {'color': 'magenta'})

    #### Manager for inference
    manager = FuseManagerDefault()
    # TODO - define the keys out of batch_dict that will be saved to a file
    output_columns = ['TODO']
    manager.infer(data_source=infer_data_source,
                  input_model_dir=paths['model_dir'],
                  checkpoint=infer_common_params['checkpoint'],
                  output_columns=output_columns,
                  output_file_name=infer_common_params['infer_filename'])


######################################
# Analyze Common Params
######################################
ANALYZE_COMMON_PARAMS = {}
ANALYZE_COMMON_PARAMS['infer_filename'] = INFER_COMMON_PARAMS['infer_filename']
ANALYZE_COMMON_PARAMS['output_filename'] = os.path.join(PATHS['analyze_dir'], 'all_metrics')
ANALYZE_COMMON_PARAMS['num_workers'] = 4
ANALYZE_COMMON_PARAMS['batch_size'] = 2


######################################
# Analyze Template
######################################
def analyze_template(paths: dict, analyze_common_params: dict):
    fuse_logger_start(output_path=None, console_verbose_level=logging.INFO)
    lgr = logging.getLogger('Fuse')
    lgr.info('Fuse Analyze', {'attrs': ['bold', 'underline']})

    # TODO - include all the gt processors required for analyzing
    gt_processors = {
        'TODO',
    }

    # metrics
    # Fuse TIP: use metric FuseMetricConfidenceInterval for computing confidence interval for metrics
    metrics = {
        # TODO : add required metrics
    }

    # create analyzer
    analyzer = FuseAnalyzerDefault()

    # run
    analyzer.analyze(gt_processors=gt_processors,
                     data_pickle_filename=analyze_common_params['infer_filename'],
                     metrics=metrics,
                     output_filename=analyze_common_params['output_filename'],
                     num_workers=analyze_common_params['num_workers'],
                     batch_size=analyze_common_params['num_workers'])


######################################
# Run
######################################
if __name__ == "__main__":
    # allocate gpus
    NUM_GPUS = 1
    # uncomment if you want to use specific gpus instead of automatically looking for free ones
    force_gpus = None  # [0]
    FuseUtilsGPU.choose_and_enable_multiple_gpus(NUM_GPUS, force_gpus=force_gpus)

    RUNNING_MODES = ['train']  # Options: 'train', 'infer', 'analyze'

    # train
    if 'train' in RUNNING_MODES:
        train_template(paths=PATHS, train_common_params=TRAIN_COMMON_PARAMS)

    # infer
    if 'infer' in RUNNING_MODES:
        infer_template(paths=PATHS, infer_common_params=INFER_COMMON_PARAMS)

    # analyze
    if 'analyze' in RUNNING_MODES:
        analyze_template(analyze_common_params=ANALYZE_COMMON_PARAMS)



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

from typing import Optional, List, Dict, Union

import torch
import pandas as pd

from fuse.data.data_source.data_source_from_list import FuseDataSourceFromList
from fuse.data.dataset.dataset_default import FuseDatasetDefault
from fuse.data.processor.processor_dataframe import FuseProcessorDataFrame


class FuseDatasetDataframe(FuseDatasetDefault):
    """
    Simple dataset, based on FuseDatasetDefault, that converts dataframe into dataset.
    """
    def __init__(self,
                 data: Optional[pd.DataFrame] = None,
                 data_pickle_filename: Optional[str] = None,
                 sample_desc_column: Optional[str] = 'descriptor',
                 columns_to_extract: Optional[List[str]] = None,
                 rename_columns: Optional[Dict[str, str]] = None,
                 columns_to_tensor: Optional[Union[List[str], Dict[str, torch.dtype]]] = None,
                 **kwargs):
        """
        :param data:  input DataFrame
        :param data_pickle_filename: path to a pickled DataFrame (possible gzipped)
        :param sample_desc_column: name of the sample descriptor column within the pickle file,
                                   if set to None.will simply use dataframe index as descriptors
        :param columns_to_extract: list of columns to extract from dataframe. When None (default) all columns are extracted
        :param rename_columns: rename columns from dataframe, when None (default) column names are kept
        :param columns_to_tensor: columns in data that should be converted into pytorch.tensor.
                        when list, all columns specified are transforms into tensors (type is decided by torch).
                        when dictionary, then each column is converted into the specified dtype.
                        When None (default) no columns are converted.
        :param kwargs: additional DatasetDefault arguments. See DatasetDefault

        """
        # create processor
        processor = FuseProcessorDataFrame(data=data,
                                           data_pickle_filename=data_pickle_filename,
                                           sample_desc_column=sample_desc_column,
                                           columns_to_extract=columns_to_extract,
                                           rename_columns=rename_columns,
                                           columns_to_tensor=columns_to_tensor)

        # extract descriptor list and create datasource
        descriptors_list = processor.get_samples_descriptors()

        data_source = FuseDataSourceFromList(descriptors_list)

        super().__init__(
            data_source=data_source,
            gt_processors=None,
            input_processors=None,
            processors=processor,
            **kwargs
        )



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import unittest
import os
import tempfile
import shutil

from fuse_examples.classification.skin_lesion.runner import TRAIN_COMMON_PARAMS, \
    INFER_COMMON_PARAMS, ANALYZE_COMMON_PARAMS, run_train, run_analyze, run_infer

from fuse.utils.utils_gpu import FuseUtilsGPU


class ClassificationSkinLesionTestCase(unittest.TestCase):

    def setUp(self):
        self.train_common_params = TRAIN_COMMON_PARAMS
        self.train_common_params['data.year'] = '2016'
        self.train_common_params['manager.train_params']['num_epochs'] = 5

        # Path to save model
        self.ROOT = tempfile.mkdtemp()
        # Path to store the data
        self.ROOT_DATA = tempfile.mkdtemp(prefix="skin_data")
        # Name of the experiment
        EXPERIMENT = 'InceptionResnetV2_2016_test'
        # Path to cache data
        self.CACHE_PATH = tempfile.mkdtemp(prefix="cache_data")
        # Name of the cached data folder
        EXPERIMENT_CACHE = 'ISIC_' + self.train_common_params['data.year']

        self.paths = {'data_dir': self.ROOT_DATA,
                      'model_dir': os.path.join(self.ROOT, EXPERIMENT, 'model_dir'),
                      'force_reset_model_dir': True,
                      'cache_dir': os.path.join(self.CACHE_PATH, EXPERIMENT_CACHE + '_cache_dir'),
                      'inference_dir': os.path.join(self.ROOT, EXPERIMENT, 'infer_dir'),
                      'analyze_dir': os.path.join(self.ROOT, EXPERIMENT, 'analyze_dir')}

        self.infer_common_params = INFER_COMMON_PARAMS
        self.infer_common_params['data.year'] = self.train_common_params['data.year']

        self.analyze_common_params = ANALYZE_COMMON_PARAMS
        self.analyze_common_params['data.year'] = self.train_common_params['data.year']

    def test_runner(self):
        num_gpus_allocated = FuseUtilsGPU.choose_and_enable_multiple_gpus(1, use_cpu_if_fail=True)
        if num_gpus_allocated == 0:
            self.train_common_params['manager.train_params']['device'] = 'cpu'

        run_train(self.paths, self.train_common_params)
        run_infer(self.paths, self.infer_common_params)
        results = run_analyze(self.paths, self.analyze_common_params)

        threshold = 0.65
        self.assertGreaterEqual(results['auc']['macro_avg'], threshold)

    def tearDown(self):
        # Delete temporary directories
        shutil.rmtree(self.ROOT)
        shutil.rmtree(self.ROOT_DATA)
        shutil.rmtree(self.CACHE_PATH)


if __name__ == '__main__':
    unittest.main()

